package edu.berkeley.sbp.misc;
import java.io.*;
import java.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.misc.*;
import edu.berkeley.sbp.tib.*;

public class RegressionTests {

    public static boolean yes = false;
    public static class MyWalker extends ReflectiveWalker {
        public String top(Object[] o) { return "top("+join(o)+")"; }
        public String str(String[] s) { String ret = ""; for(String st : s) ret += st; return ret; }
        public String join(Object[] o) { String ret = ""; for(Object st : o) ret += st; return ret; }
        public String whilex(Object s, Object y) { return "while("+s+") " + y; }
        public String seq(Object[] statements) {
            String ret = "";
            for(Object s : statements) ret += s + ";\n";
            return ret;
        }
        /*
        public String bl(String s) { return "{" + s + "}"; }
        */
    };

    public static void main(String[] s) throws Exception {
        try {
            boolean profile = false;
            if (s[0].equals("-profile")) {
                profile = true;
                String[] s2 = new String[s.length-1];
                System.arraycopy(s, 1, s2, 0, s2.length);
                s = s2;
            }

            //MetaGrammar mg0 = new MetaGrammar();
            //mg0.walk(MetaGrammar.meta);
            //System.out.println(mg0);
            Tree<String> res = new Parser(MetaGrammar.make(), CharToken.top()).parse1(new CharToken.Stream(new InputStreamReader(new FileInputStream(s[0]))));
            MetaGrammar mg = (MetaGrammar)new MetaGrammar().walk(res);
            //System.out.println(mg);
            Union meta = mg.done();
            SequenceInputStream sis = new SequenceInputStream(new FileInputStream(s[0]), new FileInputStream(s[1]));
            res = new Parser(meta, CharToken.top()).parse1(new CharToken.Stream(new InputStreamReader(sis), "parsing " + s[1] + " using " + s[0]));
            Union testcasegrammar = ((MetaGrammar)new MetaGrammar("ts").walk(res)).done("ts");
            if (testcasegrammar==null) return;
            CharToken.Stream cs = new CharToken.Stream(new InputStreamReader(new FileInputStream(s[2])), "parsing " + s[2] + " using " + s[1]);
            Parser parser = new Parser(testcasegrammar, CharToken.top());

            if (profile) {
                System.out.println("\nready...");
                System.in.read();
            }
            Forest<String> r2 = parser.parse(cs);
            if (profile) {
                System.out.println("\ndone");
                System.in.read();
                System.exit(0);
            }
            for(TestCase tc : (TestCase[])new TestCaseBuilder().walk(r2.expand1())) tc.execute();

        } catch (Throwable t) {
            System.err.println("\n\nexception thrown, class == " + t.getClass().getName());
            System.err.println(t);
            System.err.println();
            t.printStackTrace();
            System.err.println();
        }
    }

    public static class TestCase {
        private final Token.Stream inp;
        public final String input;
        public final String[] output;
        public final Union grammar;
        public TestCase(String input, String[] output, Union grammar, boolean tib) throws IOException {
            this.inp = tib
                ? new Tib(input)
                : new CharToken.Stream(new StringReader(input), input.indexOf('\n')==-1?"\""+input+"\": ":"");
            this.input = input;
            this.output = output;
            this.grammar = grammar;
        }
        public String toString() {
            String ret = "testcase {\n" + "  input \""+input+"\";\n";
            for(String s : output) ret += "  output \""+s+"\";\n";
            ret += grammar +"\n}\n";
            return ret;
        }
        public boolean execute() throws Exception {
            Forest<String> res = new Parser(grammar, CharToken.top()).parse(inp);
            Collection<Tree<String>> results = res==null ? new HashSet<Tree<String>>() : res.expand(false);
            System.out.print("\r");
            if (results.size() == 0 && output.length > 0) {
                System.out.print("\033[31m");
                System.out.println("PARSE FAILED");
                System.out.print("\033[0m");
            } else {
                System.out.println("\r                                                                                                              \r");
            }
            HashSet<String> outs = new HashSet<String>();
            if (output != null) for(String s : output) outs.add(s.trim());
            boolean bad = false;
            for (Tree<String> r : results) {
                String s = r.toString().trim();
                if (outs.contains(s)) { outs.remove(s); continue; }
                System.out.print("\033[33m");
                System.out.println("     GOT: " + s);
                bad = true;
            }
            for(String s : outs) {
                System.out.print("\033[31m");
                System.out.println("EXPECTED: " + s);
                bad = true;
            }
            if (bad) {
                System.out.print("\033[0m");
                return true;
            }             
            System.out.println("\033[32mPASS\033[0m");
            return false;
        }
    }

    public static class TestCaseBuilder extends MetaGrammar {
        public TestCase[] ts(Object o1, TestCase[] ts, Object o2) { return ts; }
        public TestCase[] ts(TestCase[] ts) { return ts; }
        public TestCase testcase(String input, String[] output, Union grammar) throws IOException {
            return new TestCase(input, output,        grammar, false); }
        public TestCase testcase(String input,                  Union grammar) throws IOException {
            return new TestCase(input, new String[0], grammar, false); }
        public TestCase tibcase(String input, String[] output, Union grammar) throws IOException {
            return new TestCase(input, output,        grammar, true); }
        public TestCase  tibcase(String input,                  Union grammar) throws IOException {
            return new TestCase(input, new String[0], grammar, true); }
        public MetaGrammar grammar(Object[] o) { return this; }
        
        public Object walk(Tree<String> tree) {
            if ("grammar".equals(tree.head())) { walkChildren(tree); return done("s"); }
            else return super.walk(tree);
        }
    }

    private static String pad(int i,String s) { return s.length() >= i ? s : pad(i-1,s)+" "; }
}

